use char_enum::{char_enum_derive::FromChar, FromChar, FromStrError, ToChar};
use std::{error::Error, fmt::Display, num::ParseIntError, str::FromStr};

use itertools::Itertools;

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, FromChar)]
pub enum IdKind {
    #[value = "u"]
    Uid,
    #[value = "g"]
    Gid,
    #[value = "b"]
    Both,
}

#[derive(Debug, Clone)]
pub struct TypedIdRangeError(String);

impl From<FromStrError> for TypedIdRangeError {
    fn from(value: FromStrError) -> Self {
        Self(value.to_string())
    }
}

impl From<ParseIdRangeError> for TypedIdRangeError {
    fn from(value: ParseIdRangeError) -> Self {
        Self(value.to_string())
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct TypedIdRange(IdKind, IdRange);

impl FromStr for TypedIdRange {
    type Err = TypedIdRangeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.split_once(":") {
            Some((kind, rest)) => Ok(Self(IdKind::from_str(kind)?, IdRange::from_str(rest)?)),
            None => Err(FromStrError::new(s.to_string()).into()),
        }
    }
}

impl Display for TypedIdRange {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}:{}", self.0, self.1)
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum ParseIdRangeError {
    InvalidArguments(usize),
    NotANumber(ParseIntError),
}

impl Display for ParseIdRangeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParseIdRangeError::InvalidArguments(args) => write!(
                f,
                "invalid number of arguments given (got {args}, expected 3)"
            ),
            ParseIdRangeError::NotANumber(err) => {
                write!(f, "could not parse argument: {err}")
            }
        }
    }
}

impl Error for ParseIdRangeError {}

impl From<ParseIntError> for ParseIdRangeError {
    fn from(value: ParseIntError) -> Self {
        Self::NotANumber(value)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct IdRange(usize, usize, usize);

impl FromStr for IdRange {
    type Err = ParseIdRangeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts: Vec<&str> = s.split(':').collect();

        let len = parts.len();
        if len > 3 {
            return Err(ParseIdRangeError::InvalidArguments(len));
        }

        let range: Vec<usize> = parts
            .iter()
            .map(|n| usize::from_str_radix(n, 10))
            .process_results(|iter| iter.collect())?;

        Ok(Self(range[0], range[1], range[2]))
    }
}

impl Display for IdRange {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}:{}:{}", self.0, self.1, self.2)
    }
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct IdMapping(Vec<IdRange>);

impl FromStr for IdMapping {
    type Err = ParseIdRangeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.split(',')
            .map(|s| IdRange::from_str(s))
            .process_results(|iter| Self(iter.collect()))
    }
}

impl Display for IdMapping {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut iter = self.0.iter();

        if let Some(head) = iter.next() {
            write!(f, "{head}")?;

            for item in iter {
                write!(f, ",{item}")?;
            }
        }

        Ok(())
    }
}
